@@ -74,6 +74,7 @@ use Data::Dumper;
 	BAD_NEGtoPOSfp0_MODE
 	BAD_NUM_DIMS
 	BAD_NUM_ELEMS
+	BAD_NUM_STRINGS
 	BAD_NUM_VARS
 	BAD_READONLY_MODE
 	BAD_REC_COUNT
@@ -85,9 +86,14 @@ use Data::Dumper;
 	BAD_VAR_NAME
 	BAD_VAR_NUM
 	BAD_zMODE
+	BADDATE_LEAPSECOND_UPDATED
+	BLOCKINGFACTOR_TOO_SMALL
+	BLOCKINGFACTOR_TOO_SMALL2
+	BLOCKINGFACTOR_TOO_LARGE
 	CANNOT_ALLOCATE_RECORDS
 	CANNOT_CHANGE
 	CANNOT_COMPRESS
+	CANNOT_INSERT_RECORDS
 	CANNOT_SPARSEARRAYS
 	CANNOT_SPARSERECORDS
 	CDF_
@@ -202,6 +208,7 @@ use Data::Dumper;
 	EPOCHx_FORMAT_MAX
 	EPOCHx_STRING_MAX
 	FORCED_PARAMETER
+	FUNCTION_NOT_SUPPORTED
 	GET_
 	GLOBAL_SCOPE
 	GLOBAL_SCOPE_ASSUMED
@@ -299,15 +306,20 @@ use Data::Dumper;
 	STAGE_CACHESIZE_
 	STATUS_TEXT_
 	STRINGDELIMITER
+	STRING_NOT_UTF8_ENCODING
 	SUN_DECODING
 	SUN_ENCODING
 	TOO_MANY_PARMS
 	TOO_MANY_VARS
-	TTT2000_0_STRING_LEN
-	TTT2000_1_STRING_LEN
-	TTT2000_2_STRING_LEN
-	TTT2000_3_STRING_LEN
-	TTT2000_4_STRING_LEN
+	TRY_TO_READ_NONSTRING_DATA
+	TT2000_0_STRING_LEN
+	TT2000_1_STRING_LEN
+	TT2000_2_STRING_LEN
+	TT2000_3_STRING_LEN
+	TT2000_4_STRING_LEN
+	TT2000_CDF_MAYNEEDUPDATE
+	TT2000_TIME_ERROR
+	TT2000_USED_OUTDATED_TABLE
 	UNKNOWN_COMPRESSION
 	UNKNOWN_SPARSENESS
 	UNSUPPORTED_OPERATION
@@ -342,6 +354,8 @@ use Data::Dumper;
 	VAX_DECODING
 	VAX_ENCODING
 	VIRTUAL_RECORD_DATA
+	ZLIB_COMPRESS_ERROR
+	ZLIB_UNCOMPRESS_ERROR
 	gENTRY_
 	gENTRY_DATASPEC_
 	gENTRY_DATATYPE_
@@ -1230,12 +1244,12 @@ sub CDFvarCreate {
         my $id = shift;
         my $varName = shift;
         my $dataType = shift;
-        my $numElements = shift;
+        my $numElems = shift;
         my $recVary = shift;
         my $dimVarysRef = shift;
         my $varNum = shift;
         my $status;
-        $status = CDFvarCreateX( $id, $varName, $dataType, $numElements, 
+        $status = CDFvarCreateX( $id, $varName, $dataType, $numElems, 
 				 $recVary, $dimVarysRef, $$varNum);
         return $status;
 }#end sub CDFvarCreate
@@ -1265,12 +1279,12 @@ sub CDFvarInquire {
         my $varNum = shift;
         my $varName = shift;
         my $dataType = shift;
-        my $numElements = shift;
+        my $numElems = shift;
         my $recVary = shift;
         my $dimVarysRef = shift;
         my $status;
         $status = CDFvarInquireX( $id, $varNum, $$varName, $$dataType, 
-				  $$numElements, $$recVary, $dimVarysRef);
+				  $$numElems, $$recVary, $dimVarysRef);
         return $status;
 }#end sub CDFvarInquire
 
@@ -1511,16 +1525,29 @@ sub toEncodeEPOCH16 {
         if ($style < 0 || $style > 4) {
          $style = 4;
         }
-        if ($style == 0) {
-            encodeEPOCH16 ($epoch, $_[2]);
-        } elsif ($style == 1) {
-            encodeEPOCH16_1 ($epoch, $_[2]);
-        } elsif ($style == 2) {
-            encodeEPOCH16_2 ($epoch, $_[2]);
-        } elsif ($style == 3) {
-            encodeEPOCH16_3 ($epoch, $_[2]);
+        my ($i, $count, $epoch16Str);
+        $count = scalar @$epoch / 2;
+        my (@epoch16, @epoch1);
+        for ($i = 0; $i < $count; $i++) {
+          $epoch1[0] = $$epoch[2*$i];
+          $epoch1[1] = $$epoch[2*$i+1];
+          if ($style == 0) {
+            encodeEPOCH16 (\@epoch1, $epoch16Str);
+          } elsif ($style == 1) {
+            encodeEPOCH16_1 (\@epoch1, $epoch16Str);
+          } elsif ($style == 2) {
+            encodeEPOCH16_2 (\@epoch1, $epoch16Str);
+          } elsif ($style == 3) {
+            encodeEPOCH16_3 (\@epoch1, $epoch16Str);
+          } else {
+            encodeEPOCH16_4 (\@epoch1, $epoch16Str);
+          }
+          $epoch16[$i] = $epoch16Str;
+        }
+        if ($count == 1) {
+          $_[2] = $epoch16[0];
         } else {
-            encodeEPOCH16_4 ($epoch, $_[2]);
+          @$stra = @epoch16;
         }
 }#end sub toEncodeEPOCH16
 
@@ -1540,15 +1567,15 @@ sub EPOCHtoUnixTime {
 sub EPOCH16toUnixTime {
         my $epoch = $_[0];
 	my $unixTime = $_[1];
-	if (ref($epoch) ne "ARRAY") {
-          my @epochs;
-          $epochs[0] = $epoch;
-          $epochs[1] = 0;
-          EPOCH16toUnixTimeX (\@epochs, $unixTime);
-          $_[1] = $times[0];
-	} else {
+        my $count;
+        $count = scalar @$epoch / 2;
+        if ($count == 1) {
+          my @unixTimes;
+          EPOCH16toUnixTimeX ($epoch, \@unixTimes);
+          $_[1] = $unixTimes[0];
+        } else {
           EPOCH16toUnixTimeX ($epoch, $unixTime);
-	}
+        }
 }#end sub EPOCH16toUnixTime
 
 sub TT2000toUnixTime {
@@ -1565,8 +1592,8 @@ sub TT2000toUnixTime {
 }#end sub TT2000toUnixTime
 
 sub UnixTimetoEPOCH {
-        my $epoch = $_[1];
 	my $unixTime = $_[0];
+        my $epoch = $_[1];
 	if (ref($unixTime) ne "ARRAY") {
           my (@epochs, @times);
           $times[0] = $unixTime;
@@ -1578,21 +1605,20 @@ sub UnixTimetoEPOCH {
 }#end sub UnixTimetoEPOCH
 
 sub UnixTimetoEPOCH16 {
+        my $unixTime = $_[0];
         my $epoch = $_[1];
-	my $unixTime = $_[0];
 	if (ref($unixTime) ne "ARRAY") {
           my @times;
           $times[0] = $unixTime;
           UnixTimetoEPOCH16X (\@times, $epoch);
-          $_[1] = \@epochs;
 	} else {
           UnixTimetoEPOCH16X ($unixTime, $epoch);
 	}
 }#end sub UnixTimetoEPOCH16
 
 sub UnixTimetoTT2000 {
+        my $unixTime = $_[0];
         my $epoch = $_[1];
-	my $unixTime = $_[0];
 	if (ref($unixTime) ne "ARRAY") {
           my (@epochs, @times);
           $times[0] = $unixTime;
@@ -1753,22 +1779,24 @@ sub CDFgetGlobalMetaData {
         my $metaEncoding = 1;
         if (scalar(@_) > 1) {
           my $haveGlobals = $args[1];
-          if (ref($haveGlobals) eq "ARRAY") {
-            if ($#$haveGlobals == -1) {
-              $enteredGlobals = -1;
-            } else {
-              @globals = @$haveGlobals;
-              $enteredGlobals = $#globals + 1;
-            }
-          } else {
-            if (!(defined $haveGlobals)) {
-              $enteredGlobals = -1;
-            } else {
-              if ($haveGlobals ne "ALL") {
-                @globals = split(',', $haveGlobals);
-                $enteredGlobals = $#globals + 1;
-              } else {
+          if ($haveGlobals) {
+            if (ref($haveGlobals) eq "ARRAY") {
+              if ($#$haveGlobals == -1) {
                 $enteredGlobals = -1;
+              } else {
+                @globals = @$haveGlobals;
+                $enteredGlobals = $#globals + 1;
+              }
+            } else {
+              if (!$haveGlobals) {
+                $enteredGlobals = -1;
+              } else {
+                if ($haveGlobals ne "ALL") {
+                  @globals = split(',', $haveGlobals);
+                  $enteredGlobals = $#globals + 1;
+                } else {
+                  $enteredGlobals = -1;
+                }
               }
             }
           }
@@ -1779,10 +1807,10 @@ sub CDFgetGlobalMetaData {
             $metaEncoding = 1;
           }
         }
-        my $id = ();
+        my $id;
         my ($toClose, $status, $cdfName, $numAttrs, $i, $scope, $attrName);
-        my %metax = ();
         $toClose = 0;
+        my %metax = ();
         if ($cdf =~ /^\d+$/) {
           $status = CDFlib(&SELECT_, &CDF_, $cdf,
                            &NULL_);
@@ -1804,7 +1832,7 @@ sub CDFgetGlobalMetaData {
         $status = CDFlib (&GET_, &CDF_NUMATTRS_, \$numAttrs,
                           &NULL_);
         if ($status != &CDF_OK) {
-          return ($status, %metax);
+          return ($status, ());
         }
         my $found;
         for ($i = 0; $i < $numAttrs; $i++) {
@@ -1853,7 +1881,7 @@ sub CDFgetGlobalMetaData {
                                  &NULL_);
                 if ($status != &CDF_OK) {
                   next;
-                } 
+                }
                 if ($dataType == &CDF_EPOCH16 || ($numElems > 1 &&
                                                   ($dataType != &CDF_CHAR &&
                                                    $dataType != &CDF_UCHAR))) {
@@ -1896,7 +1924,7 @@ sub CDFgetGlobalMetaData {
                     }
                   } else {
                     if ($dataType == &CDF_CHAR || $dataType == &CDF_UCHAR ||
-                        $numElems == 1) {
+                        $numElems == 1) { 
                       $meta2{$j} = $entry;
                     } else {
                       $meta2{$j} = \@entry2;
@@ -1920,24 +1948,101 @@ sub CDFgetGlobalMetaData {
           $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
         }
         $status = &CDF_OK;
-        return ($status, %metax);
+        if (%metax) {
+          return ($status, %metax);
+        } else {
+          return ($status, ());
+        }
 }#end sub CDFgetGlobalMetaData
 
+sub CDFgetNoEntryAttrs {
+        use strict;
+        my @args = @_;
+        my $cdf = $args[0];
+        my $id;
+        my ($toClose, $status, $numAttrs, $i, $scope, $attrName);
+        $toClose = 0;
+        if ($cdf =~ /^\d+$/) {
+          $status = CDFlib(&SELECT_, &CDF_, $cdf,
+                           &NULL_);
+          if ($status != &CDF_OK) {
+            $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                             &NULL_);
+            $toClose = 1;
+          } else {
+            $id = $cdf;
+          }
+        } else {
+          $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                           &NULL_);
+          $toClose = 1;
+        }
+        if ($status != &CDF_OK) {
+          return ($status, ());
+        }
+        $status = CDFlib (&GET_, &CDF_NUMATTRS_, \$numAttrs,
+                          &NULL_);
+        if ($status != &CDF_OK) {
+          return ($status, ());
+        }
+        my (@globals, @variables);
+        my ($numEntries, $numrEntries, $numzEntries);
+        @globals = ();
+        @variables = ();
+        for ($i = 0; $i < $numAttrs; $i++) {
+          $status = CDFlib(&SELECT_, &ATTR_, $i,
+                           &GET_, &ATTR_SCOPE_, \$scope,
+                                  &ATTR_NAME_, \$attrName,
+                           &NULL_);
+          if ($scope == &GLOBAL_SCOPE) {
+            $status = CDFlib(&GET_, &ATTR_NUMgENTRIES_, \$numEntries,
+                             &NULL_);
+            if ($numEntries == 0) {
+              push(@globals, $attrName);
+            }
+          } else {
+            $status = CDFlib(&GET_, &ATTR_NUMrENTRIES_, \$numrEntries,
+                                    &ATTR_NUMzENTRIES_, \$numzEntries,
+                             &NULL_);
+            if (($numrEntries+$numzEntries) == 0) {
+              push(@variables, $attrName);
+            }
+          }
+        }
+        if ($toClose == 1) {
+          $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
+        }
+        $status = &CDF_OK;
+        my %noEntry = ();
+        if (@globals) {
+          $noEntry{"GlobalAttritbues"} = \@globals;
+        }
+        if (@variables) {
+          $noEntry{"VariableAttributes"} = \@variables;
+        }
+        if (%noEntry) {
+          return ($status, %noEntry);
+        } else {
+          return ($status, ());
+        }
+}#end sub CDFgetNoEntryAttrs
+
 sub CDFgetVarMetaData {
         use strict;
         my @args = @_;
         my $cdf = $args[0];
         my $var = $args[1];
-        my $metaEncoding;
-        if (scalar(@_) > 2) {
+        my $metaEncoding = 1;
+        if (scalar(@_) > 2) { 
           $metaEncoding = $args[2];
-        } else {
-          $metaEncoding = 1;
+          if ($metaEncoding != 0) {
+            $metaEncoding = 1;
+          }
         }
-        my $id = ();
-        my ($toClose, $varName, $varNum, $numElems, $status, $numAttrs,$zMode);
-        $toClose = 0;
+        my $id;
+        my ($toClose, $varNum, $varName, $status, $numAttrs,$zMode);
         my %metax = ();
+        $toClose = 0;
         if ($cdf =~ /^\d+$/) {
           $status = CDFlib(&SELECT_, &CDF_, $cdf,
                            &NULL_);
@@ -1970,7 +2075,7 @@ sub CDFgetVarMetaData {
           if ($status == &CDF_OK) {
             $check++;
           }
-        } 
+        }
         if ($check == 0) {
           $varName = $var;
           $status = CDFlib(&GET_, &zVAR_NUMBER_, $varName, \$varNum,
@@ -1982,16 +2087,7 @@ sub CDFgetVarMetaData {
             return (&BAD_VAR_NAME, ());
           }
         }
-        $status = CDFlib (&SELECT_, &zVAR_, $varNum,
-                                    &zENTRY_, $varNum,
-                          &NULL_);
-        if ($status != &CDF_OK) {
-          if ($toClose == 1) {
-            $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
-          }
-          return ($status, ());
-        }
-        my ($i, $scope, $attrName, $entry, $dataType);
+        my ($i, $scope, $attrName, $entry, $dataType, $numElems);
         for ($i = 0; $i < $numAttrs; $i++) {
           my @entry2 = ();
           $status = CDFlib(&SELECT_, &ATTR_, $i,
@@ -2035,15 +2131,7 @@ sub CDFgetVarMetaData {
                   $metax{$attrName} = $epoch2;
                 } else {
                   my @epoch2;
-                  my ($jj, $ep16String);
-                  my @ep16;
-                  for ($jj = 0; $jj < $numElems; ++$jj) {
-                    $ep16[0] = $entry2[2*$jj];
-                    $ep16[1] = $entry2[2*$jj+1];
-                    toEncodeEPOCH16 (\@ep16, 4, $ep16String);
-                    $epoch2[$jj] = $ep16String;
-                  }
-#                  toEncodeEPOCH16 (\@entry2, 4, \@epoch2);
+                  toEncodeEPOCH16 (\@entry2, 4, \@epoch2);
                   $metax{$attrName} = \@epoch2;
                 }
               } elsif ($dataType == &CDF_TIME_TT2000) {
@@ -2069,7 +2157,7 @@ sub CDFgetVarMetaData {
                                                 $dataType != &CDF_UCHAR &&
                                                 $numElems > 1)) {
                 $metax{$attrName} = \@entry2;
-              } else {
+              } else { 
                 $metax{$attrName} = $entry;
               }
             }
@@ -2103,13 +2191,251 @@ sub CDFgetLIBInfo {
         return (&CDF_OK, %libInfo);
 }#end sub CDFgetLIBInfo
 
+sub CDFreadCDF {
+
+        use strict;
+        my @args = @_;
+        my $cdf = $args[0];
+        my @vars = ();
+        my $enteredVars = -1;
+        my ($toClose, $metaEncoding, $dataEncoding, $matrix);
+
+        if (scalar(@_) > 1) {
+          my $haveVars = $args[1];
+          if (ref($haveVars) eq "ARRAY") {
+            if ($#$haveVars == -1) {
+              $enteredVars = -1;
+            } else {
+              @vars = @$haveVars;
+              $enteredVars = $#vars + 1;
+            }
+          } else {
+            if (!$haveVars) {
+              $enteredVars = -1;
+            } else {
+              if ($haveVars ne "ALL") {
+                @vars = split(',', $haveVars);
+                $enteredVars = $#vars + 1;
+              } else {
+                $enteredVars = -1;
+              }
+            }
+          }
+        }
+
+        $toClose = 0;
+        $metaEncoding = 1;
+        $dataEncoding = 0;
+        $matrix = 0;
+        if (scalar(@_) > 2) {
+          $metaEncoding = $args[2];
+          if ($metaEncoding != 0) {
+            $metaEncoding = 1;
+          }
+        }
+        if (scalar(@_) > 3) {
+          $dataEncoding = $args[3];
+          if ($dataEncoding != 0) {
+            $dataEncoding = 1;
+          }
+        }
+        if (scalar(@_) > 4) {
+          $matrix = $args[4];
+          if ($matrix != 0) {
+            $matrix = 1;
+          }
+        }
+        my $id;
+        my $status;
+        my %cdfInfo;
+        if ($cdf =~ /^\d+$/) {
+          $status = CDFlib(&SELECT_, &CDF_, $cdf,
+                           &NULL_);
+          if ($status != &CDF_OK) {
+            $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                             &NULL_);
+            $toClose = 1;
+          } else {
+            $id = $cdf;
+          }
+        } else {
+          $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                           &NULL_);
+          $toClose = 1;
+        }
+        if ($status != &CDF_OK) {
+          return ($status, ());
+        }
+        my $zmodeO;
+        $status = CDFlib(&CONFIRM_, &CDF_zMODE_, \$zmodeO,
+                         &SELECT_, &CDF_zMODE_, &zMODEon2,
+                         &NULL_);
+        my (%cdfinfo, %globalAttrs, %cdfTotal, %noEntry);
+        %globalAttrs = ();
+        %noEntry = ();
+        ($status, %cdfinfo) = CDFgetCDFInfo($id);
+        if ($status == &CDF_OK) {
+          $cdfTotal{"CDFInfo"} = \%cdfinfo;
+        } else {
+          return ($status, ());
+        }
+        my ($i, $j);
+        my @dummy = ();
+        ($status, %globalAttrs) = CDFgetGlobalMetaData($id, \@dummy,
+                                                       $metaEncoding);
+        if ($status == &CDF_OK) {
+          if (%globalAttrs) {
+            $cdfTotal{"GlobalAttritbues"} = \%globalAttrs;
+          }
+        } else {
+          return ($status, ());
+        }
+        my %variables = ();
+        my $numVars;
+        $status = CDFlib(&GET_, &CDF_NUMzVARS_, \$numVars,
+                         &NULL_);
+        my $found;
+        if ($enteredVars == -1) {
+          $enteredVars = $numVars;
+          for ($i = 0; $i < $numVars; $i++) {
+            $vars[$i] = $i;
+          }
+        }
+        my ($k, $varNum, $varName);
+        for ($k = 0; $k < $enteredVars; $k++) {
+          my %varInfo = ();
+          my $varInfoP;
+          if (!($vars[$k] =~ /^\d+$/)) {
+             $varName = $vars[$k];
+             $status = CDFlib(&SELECT_, &zVAR_NAME_, $varName,
+                              &GET_, &zVAR_NUMBER_, $varName, \$varNum,
+                              &NULL_);
+             if ($status == &CDF_OK) {
+               ($status,  %varInfo) = CDFreadVar($id, $varNum, $metaEncoding,
+                                                 $dataEncoding, $matrix);
+               if ($status == &CDF_OK) {
+                 $variables{$varName} = \%varInfo;
+               } else {
+                 return ($status, ());
+               }
+             }
+          } else {
+             $varNum = $vars[$k];
+             $status = CDFlib(&SELECT_, &zVAR_, $varNum,
+                              &GET_, &zVAR_NAME_, \$varName,
+                              &NULL_);
+             if ($status == &CDF_OK) {
+               ($status,  %varInfo) = CDFreadVar($id, $varNum, $metaEncoding,
+                                                 $dataEncoding, $matrix);
+               if ($status == &CDF_OK) {
+                 $variables{$varName} = \%varInfo;
+               } else { 
+                 return ($status, ());
+               }
+             } else {
+               return ($status, ());
+             } 
+          }
+        }
+        if (%variables) {
+          $cdfTotal{"Variables"} = \%variables;
+        }
+
+        ($status,  %noEntry) = CDFgetNoEntryAttrs($id);
+        if ($status == &CDF_OK && %noEntry) {
+          $cdfTotal{"NoEntryAttrs"} = \%noEntry;
+        }
+
+        if ($zmodeO != &zMODEon2) {
+          $status = CDFlib(&SELECT_, &CDF_zMODE_, $zmodeO,
+                           &NULL_);
+        }
+        if ($toClose == 1) {
+          $status = CDFlib(&CLOSE_, &CDF_, 
+                           &NULL_);
+        }
+        return (&CDF_OK, %cdfTotal); 
+} # end sub CDFreadCDF
+
+sub CDFreadVar {
+        my @args = @_;
+        my $cdf = $args[0];
+        my @vars = ();
+        my $enteredVars = 0;
+        my ($toClose, $metaEncoding, $dataEncoding, $matrix);
+        $varID = $args[1];
+        $toClose = 0;
+        $metaEncoding = 1;
+        $dataEncoding = 0;
+        $matrix = 0;
+        if (scalar(@_) > 2) {
+          $metaEncoding = $args[2];
+          if ($metaEncoding != 0) {
+            $metaEncoding = 1;
+          }
+        }
+        if (scalar(@_) > 3) {
+          $dataEncoding = $args[3];
+          if ($dataEncoding != 0) {
+            $dataEncoding = 1;
+          }
+        }
+        if (scalar(@_) > 4) {
+          $matrix = $args[4];
+          if ($matrix != 0) {
+            $matrix = 1;
+          }
+        }
+        my $id;
+        my $status;
+        if ($cdf =~ /^\d+$/) {
+          $status = CDFlib(&SELECT_, &CDF_, $cdf,
+                           &NULL_);
+          if ($status != &CDF_OK) {
+            $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                             &NULL_);
+            $toClose = 1;
+          } else {
+            $id = $cdf;
+          }
+        } else {
+          $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                           &NULL_);
+          $toClose = 1;
+        }
+        if ($status != &CDF_OK) {
+          return ($status, ());
+        }
+        my %varDef = ();
+        my %varAttrs = ();
+        my @varData = ();
+        ($status, %varDef) = CDFgetVarInfo($id, $varID);
+        ($status, %varAttrs) = CDFgetVarMetaData($id, $varID,
+                                                 $metaEncoding);
+        ($status, @varData) = CDFgetVarAllData($id, $varID, $dataEncoding,
+                                               $matrix);
+        my %varInfo;
+        $varInfo{"VarInfo"} = \%varDef;
+        if (%varAttrs) {
+          $varInfo{"VarMetaData"} = \%varAttrs;
+        }
+        if (@varData) {
+          $varInfo{"VarData"} = \@varData;
+        }
+        if ($toClose == 1) {
+          $status = CDFlib(&CLOSE_, &CDF_,
+                           &NULL_);
+        }
+        return ($status, %varInfo);
+} # end sub CDFreadVar
+
 sub CDFgetCDFInfo {
         use strict;
         my $cdf = shift;
-        my $id = ();
-        my ($toClose, $numrVars, $numzVars, $status, $numgAttrs, $numvAttrs);
+        my $id;
+        my ($numrVars, $numzVars, $status, $numgAttrs, $numvAttrs);
         my ($majority, $format, $encoding, $release, $version, $increment);
-        my ($lastleap, $cdfName);
+        my ($toClose, $lastleap, $cdfName);
         my %cdfInfo;
         $toClose = 0;
         if ($cdf =~ /^\d+$/) {
@@ -2146,12 +2472,15 @@ sub CDFgetCDFInfo {
         if ($toClose == 1) {
           $cdfInfo{"NAME"} = $cdf;
         }
-        $cdfInfo{"FORMAT"} = getFormat($format);
-        $cdfInfo{"MAJORITY"} = getMajority($majority);
-        $cdfInfo{"ENCODING"} = getEncoding($encoding);
+#       $cdfInfo{"FORMAT"} = getFormat($format);
+        $cdfInfo{"FORMAT"} = $format;
+#       $cdfInfo{"MAJORITY"} = getMajority($majority);
+        $cdfInfo{"MAJORITY"} = $majority;
+#       $cdfInfo{"ENCODING"} = getEncoding($encoding);
+        $cdfInfo{"ENCODING"} = $encoding;
         $cdfInfo{"CDF_VERSION"} = $version.".".$release.".".$increment;
         $cdfInfo{"BASED_LEAPSECOND_LAST_UPDATED"} = $lastleap;
-        $cdfInfo{"NUMrVARS"} = $numrVars;
+#        $cdfInfo{"NUMrVARS"} = $numrVars;
         $cdfInfo{"NUMzVARS"} = $numzVars;
         $cdfInfo{"NUMgATTRS"} = $numgAttrs;
         $cdfInfo{"NUMvATTRS"} = $numvAttrs;
@@ -2166,9 +2495,9 @@ sub CDFgetVarInfo {
         my @args = @_;
         my $cdf = $args[0];
         my $var = $args[1];
-        my $id = ();
-        my ($toClose, $varNum, $status, $dataType, $numElems);
-        my ($numDims, $recVary, $maxRec, $varName, $zMode);
+        my $id;
+        my ($toClose, $varNum, $status, $dataType, $numElems, $sparse);
+        my ($numDims, $recVary, $maxRec, $varName, $zMode, $numRecs);
         my (@dimSizes, @dimVarys);
         my %varInfo;
         $toClose = 0;
@@ -2226,6 +2555,8 @@ sub CDFgetVarInfo {
                                  &zVAR_DIMVARYS_, \@dimVarys,
                                  &zVAR_RECVARY_, \$recVary,
                                  &zVAR_MAXREC_, \$maxRec,
+                                 &zVAR_NUMRECS_, \$numRecs,
+                                 &zVAR_SPARSERECORDS_, \$sparse,
                           &NULL_);
         if ($status != &CDF_OK) {
           if ($toClose == 1) {
@@ -2233,11 +2564,12 @@ sub CDFgetVarInfo {
           }
           return ($status, ());
         }
-        $varInfo{"DataType"} = getStringDataType($dataType);
+#       $varInfo{"DataType"} = getStringDataType($dataType);
+        $varInfo{"DataType"} = $dataType;
         $varInfo{"NumElems"} = $numElems;
         $varInfo{"NumDims"} = $numDims;
         if ($numDims > 0) {
-          my $i; my %indices;
+          my $i;
           for ($i = 0; $i < $numDims; $i++) {
             $varInfo{"DimSizes"}{$i} = $dimSizes[$i];
           }
@@ -2247,7 +2579,12 @@ sub CDFgetVarInfo {
         }
         $varInfo{"RecVary"} = $recVary==0?"False":"True";
         $varInfo{"VarName"} = $varName;
-        $varInfo{"MaxRec"} = $maxRec;
+        $varInfo{"NumWrittenRecs"} = $numRecs;
+        if ($numRecs > 0 && $sparse != &NO_SPARSERECORDS) {
+          $varInfo{"MaxRec"} = $maxRec;
+#         $varInfo{"SparseRecords"} = getSparseness($sparse);
+          $varInfo{"SparseRecords"} = $sparse;
+        }
         if ($toClose == 1) {
           $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
         } else {
@@ -2257,26 +2594,17 @@ sub CDFgetVarInfo {
         return (&CDF_OK, %varInfo);
 }#end sub CDFgetVarInfo
 
-sub CDFgetVarAllData {
+sub CDFgetVarAttrs {
         use strict;
         my @args = @_;
         my $cdf = $args[0];
         my $var = $args[1];
-        my $id = ();
-        my ($matrix, $varNum, $status, $maxRec, $numDims, $varName, $dataType);
-        my (@dimSizes, @dimIndices, @dimIntervals);
-        my ($toClose, $zMode);
-        my (@values, @outData);
+        my $id;
+        my ($varNum, $status, $dataType, $numElems);
+        my ($toClose, $numDims, $recVary, $maxRec, $varName, $zMode);
+        my (@dimSizes, @dimVarys);
+        my %varInfo;
         $toClose = 0;
-        $matrix = 0;
-        my $dataEncoding = 0;
-        if (scalar(@_) > 2) {
-          $dataEncoding = $args[2];
-        }
-        if (scalar(@_) > 3) {
-          $matrix = $args[3];
-        }
-
         if ($cdf =~ /^\d+$/) {
           $status = CDFlib(&SELECT_, &CDF_, $cdf,
                            &NULL_);
@@ -2323,6 +2651,101 @@ sub CDFgetVarAllData {
             return (&BAD_VAR_NAME, ());
           }
         }
+        my $numAttrs;
+        $status = CDFlib (&GET_, &CDF_NUMvATTRS_, \$numAttrs,
+                          &NULL_);
+        if ($status != &CDF_OK) {
+          if ($toClose == 1) {
+            $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
+          }
+          return ($status, ());
+        }
+        my %varAttrs;
+        for ($i = 0; $i < $numAttrs; $i++) {
+          $status = CDFlib (&SELECT_, &vATTR_, $i,
+                                      &zENTRY_, $vaNum,
+                            &GET_, &zENTRY_DATA_, \$entry,
+                            &NULL_);
+          if ($status == CDF_OK) {
+            $varAttrs[$attrName] = $entry;
+          }
+        }
+        if ($toClose == 1) {
+          $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
+        } else {
+          $status = CDFlib (&SELECT_, &CDF_zMODE_, $zMode,
+                            &NULL_);
+        }
+        return (&CDF_OK, %varAttrs);
+}#end sub CDFgetVarInfo
+
+sub CDFgetVarAllData {
+        use strict;
+        my @args = @_;
+        my $cdf = $args[0];
+        my $var = $args[1];
+        my ($toClose, $matrix, $dataEncoding);
+        $toClose = 0;
+        $matrix = 0;
+        $dataEncoding = 0;
+        if (scalar(@_) > 2) { 
+          $dataEncoding = $args[2];
+        }
+        if (scalar(@_) > 3) {
+          $matrix = $args[3];
+        }
+
+        my $id;
+        my ($varNum, $status, $maxRec, $numDims, $varName, $dataType);
+        my (@dimSizes, @dimIndices, @dimIntervals);
+        my $zMode;
+        my (@values, @outData);
+        if ($cdf =~ /^\d+$/) {
+          $status = CDFlib(&SELECT_, &CDF_, $cdf,
+                           &NULL_);
+          if ($status != &CDF_OK) {
+            $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                             &NULL_);
+            $toClose = 1;
+          }
+        } else {
+          $status = CDFlib(&OPEN_, &CDF_, $cdf, \$id,
+                           &NULL_);
+          $toClose = 1;
+        }
+        if ($status != &CDF_OK) {
+          return ($status, ());
+        }
+        $status = CDFlib(&CONFIRM_, &CDF_zMODE_, \$zMode,
+                         &SELECT_, &CDF_zMODE_, &zMODEon2,
+                         &NULL_);
+        if ($status != &CDF_OK) {
+          if ($toClose == 1) {
+            $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
+          }
+          return ($status, ());
+        }
+        my $check = 0;
+        if ($var =~ /^\d+$/) {
+          $varNum = $var;
+          $status = CDFlib(&SELECT_, &zVAR_, $varNum,
+                           &GET_, &zVAR_NAME_, $varName,
+                           &NULL_);
+          if ($status == &CDF_OK) {
+            $check++;
+          }
+        }
+        if ($check == 0) {
+          $varName = $var;
+          $status = CDFlib(&GET_, &zVAR_NUMBER_, $varName, \$varNum,
+                           &NULL_);
+          if ($status != &CDF_OK) {
+            if ($toClose == 1) {
+              $status = CDFlib(&CLOSE_, &CDF_, &NULL_);
+            }
+            return (&BAD_VAR_NAME, ());
+          }
+        }
         $status = CDFlib (&SELECT_, &zVAR_, $varNum,
                           &GET_, &zVAR_DIMSIZES_, \@dimSizes,
                                  &zVAR_MAXREC_, \$maxRec,
@@ -2348,29 +2771,17 @@ sub CDFgetVarAllData {
                                       &zVAR_DIMINTERVALS_, \@dimIntervals,
                             &NULL_);
           $status = CDFlib2(&GET_, &zVAR_HYPERDATA_,\@values);
-          if ($dataEncoding == 1) {
-            if ($dataType == &CDF_EPOCH) {
-              toEncodeEPOCH(\@values, 4, \@outData);
-            } elsif ($dataType == &CDF_TIME_TT2000) {
-              toEncodeTT2000(\@values, 3, \@outData);
-            } elsif ($dataType == &CDF_EPOCH16) {
-              my @ep16;
-              my $ep16String;
-              for ($i = 0; $i <= $maxRec; ++$i) {
-                $ep16[0] = $values[2*$i];
-                $ep16[1] = $values[2*$i+1];
-                toEncodeEPOCH16(\@ep16, 3, $ep16String);
-                $outData[$i] = $ep16String;
-              }
-            } else {
-              @outData = @values;
-            }
+          if ($dataType == &CDF_EPOCH) {
+            toEncodeEPOCH(\@values, 4, \@outData);
+          } elsif ($dataType == &CDF_TIME_TT2000) {
+            toEncodeTT2000(\@values, 3, \@outData);
+          } elsif ($dataType == &CDF_EPOCH16) {
+            toEncodeEPOCH16(\@values, 4, \@outData);
           } else {
             @outData = @values;
           }
-
           if ($status >= &CDF_OK && $matrix == 1) {
-            if ($dataType == &CDF_EPOCH16) {
+            if ($dataType == &CDF_EPOCH16 && $dataEncoding != 1) {
               if ($numDims == 0) {
                 $dimSizes[0] = 2;
               } else {
@@ -2383,8 +2794,8 @@ sub CDFgetVarAllData {
               if ($numRecs > 1) {
                 unshift @dimSizes, $numRecs;
               }
-              my $matrix = make_matrix(\@dimSizes, \@outData);
-              @outData = @$matrix;
+              my $matrixOut = make_matrix(\@dimSizes, \@outData);
+              @outData = @$matrixOut;
             }
           }
         } else {
@@ -2476,4 +2887,9 @@ sub getFormat {
   return $cFormat[$format];
 }
 
+sub getSparseness {
+  my $sparse = shift;
+  my @cSparse = ("NO_SPARSE", "PAD_SPARSE", "PREV_SPARSE");
+  return $cSparse[$sparse];
+}
 
